"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var create_error_1 = require("../utils/create-error");
var merge_queries_1 = require("../utils/merge-queries");
var run_query_against_cache_1 = require("../utils/run-query-against-cache");
var not_equal_1 = require("../utils/not-equal");
var merge_cache_1 = require("../utils/merge-cache");
var network_layer_1 = require("../network-layer");
exports.NetworkLayer = network_layer_1.default;
var request_batcher_1 = require("./request-batcher");
var mutation_1 = require("./mutation");
var optimistic_1 = require("./optimistic");
var OptimisticValueStore_1 = require("./OptimisticValueStore");
var NodeID_1 = require("../types/NodeID");
exports.createNodeID = NodeID_1.createNodeID;
var query_1 = require("../typed-helpers/query");
var Client = /** @class */ (function () {
    function Client(options) {
        if (options === void 0) { options = {}; }
        this._optimisticValueStore = new OptimisticValueStore_1.default();
        this._queries = [];
        this._queriesCount = [];
        this._updateHandlers = [];
        this._networkErrorHandlers = [];
        this._mutationErrorHandlers = [];
        this._queueRequestHandlers = [];
        this._successfulResponseHandlers = [];
        this._optimisticUpdaters = {};
        this._options = options;
        var serverPreparation = options.serverPreparation !== undefined
            ? options.serverPreparation
            : typeof BICYCLE_SERVER_PREPARATION !== 'undefined'
                ? BICYCLE_SERVER_PREPARATION
                : undefined;
        this._cache =
            (serverPreparation && serverPreparation.c) ||
                { Root: { root: {} } };
        this._optimisticCache = this._cache;
        this._request = new request_batcher_1.default(options.networkLayer === undefined
            ? new network_layer_1.default()
            : options.networkLayer, serverPreparation, this);
    }
    Client.prototype._onUpdate = function (fn) {
        this._updateHandlers.push(fn);
    };
    Client.prototype._offUpdate = function (fn) {
        this._updateHandlers.splice(this._updateHandlers.indexOf(fn), 1);
    };
    // called by RequestBatcher, these errors are always retried and are usually temporary
    Client.prototype._handleNetworkError = function (err) {
        err.code = 'NETWORK_ERROR';
        setTimeout(function () {
            throw err;
        }, 0);
        this._networkErrorHandlers.forEach(function (handler) {
            handler(err);
        });
    };
    // called by RequestBatcher, these errors are not retried
    Client.prototype._handleMutationError = function (err) {
        setTimeout(function () {
            throw err;
        }, 0);
        this._mutationErrorHandlers.forEach(function (handler) {
            handler(err);
        });
    };
    Client.prototype._handleQueueRequest = function () {
        this._queueRequestHandlers.forEach(function (handler) {
            handler();
        });
    };
    Client.prototype._handleSuccessfulResponse = function (pendingMutations) {
        this._successfulResponseHandlers.forEach(function (handler) {
            handler(pendingMutations);
        });
    };
    Client.prototype._handleNewSession = function (data) {
        this._cache = data;
        this._updateOptimistCache();
    };
    // called by RequestBatcher when there is new data for the cache or the list of pending mutations changes
    Client.prototype._handleUpdate = function (data) {
        if (data) {
            this._cache = merge_cache_1.default(this._cache, data);
        }
        this._updateOptimistCache();
    };
    Client.prototype._updateOptimistCache = function () {
        var pendingMutations = this._request.getPendingMutations();
        if (pendingMutations.length) {
            var result = {};
            var cache_1 = new optimistic_1.BaseCache(NodeID_1.createNodeID('Root', 'root'), this._cache, result);
            pendingMutations.forEach(function (mutation) { return mutation.applyOptimistic(cache_1); });
            this._optimisticCache = merge_cache_1.default(this._cache, result);
        }
        else {
            this._optimisticCache = this._cache;
        }
        this._updateHandlers.forEach(function (handler) {
            handler();
        });
    };
    Client.prototype._getQuery = function () {
        if (this._queries.length === 0)
            return {};
        return merge_queries_1.default.apply(void 0, this._queries);
    };
    Client.prototype._updateQuery = function () {
        this._request.updateQuery(this._getQuery());
    };
    Client.prototype._addQuery = function (query) {
        var i = this._queries.indexOf(query);
        if (i !== -1) {
            this._queriesCount[i]++;
            return;
        }
        this._queries.push(query);
        this._queriesCount.push(1);
        this._updateQuery();
    };
    Client.prototype._removeQuery = function (query) {
        var i = this._queries.indexOf(query);
        if (i === -1) {
            console.warn('You attempted to remove a query that does not exist.');
            return;
        }
        this._queriesCount[i]--;
        if (this._queriesCount[i] !== 0)
            return;
        this._queries.splice(i, 1);
        this._queriesCount.splice(i, 1);
        this._updateQuery();
    };
    Client.prototype.queryCache = function (query) {
        return run_query_against_cache_1.default(this._optimisticCache, query instanceof query_1.BaseRootQuery ? query._query : query);
    };
    Client.prototype.query = function (query) {
        var _this = this;
        var q = query instanceof query_1.BaseRootQuery ? query._query : query;
        return new Promise(function (resolve, reject) {
            var subscription = null;
            var done = false;
            subscription = _this.subscribe(q, function (result, loaded, errors) {
                if (errors.length) {
                    var err = create_error_1.default('Error fetching data for query:\n' + errors.join('\n'), {
                        code: 'BICYCLE_QUERY_ERROR',
                        query: q,
                        errors: errors,
                        result: result,
                    });
                    reject(err);
                    done = true;
                    if (subscription)
                        subscription.unsubscribe();
                }
                else if (loaded) {
                    resolve(result);
                    done = true;
                    if (subscription)
                        subscription.unsubscribe();
                }
            });
            if (done)
                subscription.unsubscribe();
        });
    };
    Client.prototype.defineOptimisticUpdaters = function (updates) {
        var _this = this;
        var u = updates;
        Object.keys(u).forEach(function (t) {
            if (!_this._optimisticUpdaters[t])
                _this._optimisticUpdaters[t] = {};
            Object.keys(u[t]).forEach(function (k) {
                _this._optimisticUpdaters[t][k] = u[t][k];
            });
        });
    };
    Client.prototype.update = function (method, args, optimisticUpdate) {
        // TODO: share a single "Mutation" class
        var m = method instanceof query_1.Mutation ? method._name : method;
        var a = method instanceof query_1.Mutation ? method._args : args;
        var o = method instanceof query_1.Mutation
            ? method._optimisticUpdate
            : optimisticUpdate;
        if (!o) {
            var split = m.split('.');
            o =
                this._optimisticUpdaters[split[0]] &&
                    this._optimisticUpdaters[split[0]][split[1]];
        }
        return this._request.runMutation(new mutation_1.default(m, a, o, this._optimisticValueStore));
    };
    Client.prototype.subscribe = function (query, fn) {
        var _this = this;
        var q = query instanceof query_1.BaseRootQuery ? query._query : query;
        var lastValue = null;
        var onUpdate = function () {
            var nextValue = _this.queryCache(q);
            if (lastValue === null || not_equal_1.default(lastValue, nextValue)) {
                lastValue = nextValue;
                fn(nextValue.result, nextValue.loaded, nextValue.errors, nextValue.errorDetails);
            }
        };
        this._onUpdate(onUpdate);
        onUpdate();
        this._addQuery(q);
        return {
            unsubscribe: function () {
                _this._offUpdate(onUpdate);
                if (_this._options.cacheTimeout) {
                    setTimeout(function () { return _this._removeQuery(q); }, _this._options.cacheTimeout);
                }
                else {
                    _this._removeQuery(q);
                }
            },
        };
    };
    Client.prototype.subscribeToNetworkErrors = function (fn) {
        var _this = this;
        this._networkErrorHandlers.push(fn);
        return {
            unsubscribe: function () {
                _this._networkErrorHandlers.splice(_this._networkErrorHandlers.indexOf(fn), 1);
            },
        };
    };
    Client.prototype.subscribeToMutationErrors = function (fn) {
        var _this = this;
        this._mutationErrorHandlers.push(fn);
        return {
            unsubscribe: function () {
                _this._mutationErrorHandlers.splice(_this._mutationErrorHandlers.indexOf(fn), 1);
            },
        };
    };
    Client.prototype.subscribeToQueueRequest = function (fn) {
        var _this = this;
        this._queueRequestHandlers.push(fn);
        return {
            unsubscribe: function () {
                _this._queueRequestHandlers.splice(_this._queueRequestHandlers.indexOf(fn), 1);
            },
        };
    };
    Client.prototype.subscribeToSuccessfulResponse = function (fn) {
        var _this = this;
        this._successfulResponseHandlers.push(fn);
        return {
            unsubscribe: function () {
                _this._successfulResponseHandlers.splice(_this._successfulResponseHandlers.indexOf(fn), 1);
            },
        };
    };
    return Client;
}());
exports.default = Client;
//# sourceMappingURL=index.js.map