"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var throat = require("throat");
var ms = require("ms");
var suggest_match_1 = require("../utils/suggest-match");
var create_error_1 = require("../utils/create-error");
var error_reporting_1 = require("../error-reporting");
var resolveField_1 = require("./resolveField");
var runQuery_1 = require("./runQuery");
var runMutation_1 = require("./runMutation");
var IS_PERFORMANCE_MONITORING = process.argv.indexOf('--monitor-bicycle-performance') !== -1;
if (IS_PERFORMANCE_MONITORING) {
    console.log('Bicycle performance monitoring is enabled.');
    console.log('Please be aware that this will make all queries much slower.');
}
var NS_PER_MS = 1e6;
function enablePerformanceMonitoring() {
    IS_PERFORMANCE_MONITORING = true;
}
exports.enablePerformanceMonitoring = enablePerformanceMonitoring;
var lock = throat(1);
function runQuery(query, _a) {
    var schema = _a.schema, logging = _a.logging, context = _a.context;
    var result = {};
    var qCtx = {
        schema: schema,
        context: context,
        logging: logging,
        result: result,
        startedQueries: {},
    };
    logging.onQueryStart({ query: query, context: context });
    if (IS_PERFORMANCE_MONITORING) {
        return lock(function () {
            var start = Date.now();
            resolveField_1.startMonitoringPerformance();
            return runQuery_1.default(schema.Root, context, query, qCtx).then(function () {
                var _a = resolveField_1.stopMonitoringPerformance(), count = _a.count, timings = _a.timings;
                var end = Date.now();
                console.log('Query completed in ' + ms(end - start));
                console.log('');
                console.log('Fields that took over 10ms to resolve:');
                console.log('');
                Object.keys(timings)
                    .sort(function (a, b) {
                    return timings[a] - timings[b];
                })
                    .forEach(function (name) {
                    if (timings[name] > 10 * NS_PER_MS) {
                        // although we compute timing in nanoseconds, this is to account for having potentially
                        // thousands of calls, each less than a millisecond.  We actually report at 1ms accuracy
                        // because everything less than that is likely just random noise.
                        console.log(" * " + name + " - " + ms(Math.round(timings[name] / NS_PER_MS)) + " (" + count[name] + " call" + (count[name] !== 1 ? 's' : '') + ")");
                    }
                });
                console.log('');
                console.log('Note that this is the **total** time to reolve the fields,');
                console.log('so you may be seeing some fields that are requested only');
                console.log('once, but are very slow, and some fields that are quick,');
                console.log('but are requested thousands of times.');
                if (logging) {
                    return Promise.resolve(logging.onQueryEnd({ query: query, cacheResult: result, context: context })).then(function () { return result; });
                }
                return result;
            });
        });
    }
    return runQuery_1.default(schema.Root, context, query, qCtx).then(function () {
        logging.onQueryEnd({ query: query, cacheResult: result, context: context });
        return result;
    });
}
exports.runQuery = runQuery;
function runMutation(mutation, mCtx) {
    return Promise.resolve(mCtx.logging.onMutationStart({ mutation: mutation, context: mCtx.context }))
        .then(function () {
        var _a = mutation.method.split('.'), typeName = _a[0], mutationName = _a[1];
        var args = mutation.args;
        var Type = mCtx.schema[typeName];
        if (!Type) {
            throw create_error_1.default('Unrecognised type for mutation: ' + typeName, {
                exposeProd: true,
                code: 'INVALID_MUTATION',
                data: { typeName: typeName, mutationName: mutationName, args: args },
            });
        }
        if (Type.kind !== 'NodeType') {
            throw create_error_1.default("The type " + typeName + " is not an Object, so you cannot call the " + mutationName + " mutation on it.", {
                exposeProd: true,
                code: 'INVALID_MUTATION',
                data: { typeName: typeName, mutationName: mutationName, args: args },
            });
        }
        if (!Type.mutations) {
            throw create_error_1.default("The type " + typeName + " does not define any mutations.", {
                exposeProd: true,
                code: 'INVALID_MUTATION',
                data: { typeName: typeName, mutationName: mutationName, args: args },
            });
        }
        var method = Type.mutations[mutationName];
        if (!method) {
            var suggestion = suggest_match_1.default(Object.keys(Type.mutations), mutationName);
            throw create_error_1.default("The type " + typeName + " does not define a mutation " + mutationName + suggestion, {
                exposeProd: true,
                code: 'INVALID_MUTATION',
                data: { typeName: typeName, mutationName: mutationName, args: args, suggestion: suggestion },
            });
        }
        return runMutation_1.default(method, mutation.args, mCtx);
    })
        .then(null, function (err) {
        var result = process.env.NODE_ENV === 'production' && !err.exposeProd
            ? {
                message: 'An unexpected error was encountered when running ' +
                    mutation.method +
                    ' (if you are the developer of this app, you can set "NODE_ENV" to "development" to expose the full error)',
                data: {},
                code: 'PRODUCTION_ERROR',
            }
            : {
                message: err.message + ' while running ' + mutation.method,
                data: err.data || {},
                code: err.code,
            };
        err.message += ' while running ' + mutation.method;
        error_reporting_1.default(err, mCtx.logging);
        return { s: false, v: result };
    })
        .then(function (result) {
        return Promise.resolve(mCtx.logging.onMutationEnd({ mutation: mutation, result: result, context: mCtx.context })).then(function () { return result; });
    });
}
exports.runMutation = runMutation;
//# sourceMappingURL=index.js.map