"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var suggest_match_1 = require("../utils/suggest-match");
var validate_1 = require("./validate");
var resolveFieldResult_1 = require("./resolveFieldResult");
var ErrorResult_1 = require("../types/ErrorResult");
var legacyArgParser_1 = require("./legacyArgParser");
var NS_PER_SEC = 1e9;
var NS_PER_MS = 1e6;
var EMPTY_OBJECT = {};
var isPerformanceMonitoring = false;
var timings = EMPTY_OBJECT;
var count = EMPTY_OBJECT;
function startMonitoringPerformance() {
    isPerformanceMonitoring = true;
    return { timings: (timings = {}), count: (count = {}) };
}
exports.startMonitoringPerformance = startMonitoringPerformance;
function stopMonitoringPerformance() {
    var oldTimings = timings;
    var oldCount = count;
    isPerformanceMonitoring = false;
    timings = EMPTY_OBJECT;
    count = EMPTY_OBJECT;
    return { timings: oldTimings, count: oldCount };
}
exports.stopMonitoringPerformance = stopMonitoringPerformance;
var waitingFields = new Map();
var queue = [];
var running = false;
var highPrecisionTiming = typeof process.hrtime === 'function';
function runTiming() {
    running = true;
    var start = highPrecisionTiming ? process.hrtime() : Date.now();
    var id = queue.pop();
    if (!id) {
        return;
    }
    var fns = waitingFields.get(id) || [];
    waitingFields.delete(id);
    return Promise.all(fns.map(function (f) { return f(); })).then(function () {
        var durationDiff = highPrecisionTiming
            ? process.hrtime(start)
            : Date.now() - start;
        var durationNanoseconds = highPrecisionTiming
            ? durationDiff[0] * NS_PER_SEC + durationDiff[1]
            : durationDiff * NS_PER_MS;
        if (typeof timings[id] !== 'number') {
            timings[id] = 0;
        }
        if (typeof count[id] !== 'number') {
            count[id] = 0;
        }
        timings[id] += durationNanoseconds;
        count[id] += fns.length;
        if (queue.length) {
            runTiming();
        }
        else {
            running = false;
        }
    });
}
function time(fn, id) {
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return new Promise(function (resolve, reject) {
            if (!waitingFields.has(id)) {
                waitingFields.set(id, []);
                queue.push(id);
            }
            (waitingFields.get(id) || []).push(function () {
                return Promise.resolve(null)
                    .then(function () { return fn.apply(void 0, args); })
                    .then(resolve, reject);
            });
            if (!running) {
                runTiming();
            }
        });
    };
}
/**
 * Resolve a single field on a node to a value to be returned to the client
 */
function resolveField(type, value, name, subQuery, qCtx) {
    var fname = name.split('(')[0];
    var args = name.indexOf('(') !== -1
        ? name
            .split('(')
            .slice(1)
            .join('(')
            .replace(/\)$/, '')
        : undefined;
    var parsedArg = undefined;
    if (args) {
        if (/^\s*[a-zA-Z0-9]+\s*\:/.test(args)) {
            parsedArg = legacyArgParser_1.default('(' + args + ')');
        }
        else {
            parsedArg = JSON.parse(args);
        }
    }
    var field = type.fields[fname];
    if (!field) {
        var suggestion = suggest_match_1.default(Object.keys(type.fields), fname);
        return Promise.resolve(ErrorResult_1.createErrorResult("Field \"" + fname + "\" does not exist on type \"" + type.name + "\"" + suggestion, undefined, 'MISSING_FIELD_NAME'));
    }
    if (field.kind === "FieldProperty" /* FieldProperty */) {
        return resolveFieldResult_1.default(field.resultType, value[fname], subQuery, qCtx).then(function (result) { return (result === undefined ? null : result); });
    }
    return Promise.resolve(null)
        .then(function () {
        if (typeof field.resolve !== 'function') {
            return ErrorResult_1.createErrorResult("Expected " + type.name + "." + fname + ".resolve to be a function.", undefined, 'INVALID_SCHEMA');
        }
        var resolveField = field.resolve;
        if (isPerformanceMonitoring) {
            resolveField = time(resolveField, type.name + "." + fname);
        }
        validate_1.validateArg(field.argType, parsedArg, qCtx.schema);
        return Promise.resolve(field.auth === 'public' ||
            field.auth(value, parsedArg, qCtx.context, subQuery, qCtx)).then(function (hasAuth) {
            if (!hasAuth) {
                return ErrorResult_1.createErrorResult("Auth failed for " + type.name + "." + fname + ".", undefined, 'AUTH_FAILED');
            }
            return resolveField(value, parsedArg, qCtx.context, subQuery, {
                fieldName: fname,
                subQuery: subQuery,
                schema: qCtx.schema,
                context: qCtx.context,
                result: qCtx.result,
                logging: qCtx.logging,
                startedQueries: qCtx.startedQueries,
            });
        });
    })
        .then(function (value) {
        if (ErrorResult_1.isErrorResult(value))
            return value;
        return resolveFieldResult_1.default(field.resultType, value, subQuery, qCtx);
    })
        .then(function (value) { return (value === undefined ? null : value); });
}
exports.default = resolveField;
//# sourceMappingURL=resolveField.js.map