"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var type_name_from_value_1 = require("../utils/type-name-from-value");
var resolveField_1 = require("./resolveField");
var error_reporting_1 = require("../error-reporting");
var ErrorResult_1 = require("../types/ErrorResult");
var NodeID_1 = require("../types/NodeID");
var is_cached_1 = require("../utils/is-cached");
function getErrorObject(err, context, logging) {
    var result = process.env.NODE_ENV === 'production' && !err.exposeProd
        ? ErrorResult_1.createErrorResult('An unexpected error was encountered ' +
            context +
            ' (if you are the developer of this app, you can set "NODE_ENV" to "development" to expose the full error)', {}, 'PRODUCTION_ERROR')
        : ErrorResult_1.createErrorResult(err.message + ' ' + context, err.data || {}, err.code);
    err.message += ' ' + context;
    error_reporting_1.default(err, logging);
    return result;
}
function makeMutableCopy(query) {
    var result = {};
    Object.keys(query).forEach(function (key) {
        var q = query[key];
        if (q === true) {
            result[key] = true;
        }
        else {
            result[key] = makeMutableCopy(q);
        }
    });
    return result;
}
function subtractQueryPart(existingQuery, newQuery) {
    var result = {};
    var added = false;
    Object.keys(newQuery).forEach(function (key) {
        var existingPart = existingQuery[key];
        var newPart = newQuery[key];
        if (!existingPart) {
            existingQuery[key] = newPart === true ? true : makeMutableCopy(newPart);
            result[key] = newPart;
            added = true;
            return;
        }
        if (existingPart === true || newPart === true) {
            return;
        }
        var diff = subtractQueryPart(existingPart, newPart);
        if (diff) {
            result[key] = diff;
            added = true;
            return;
        }
    });
    return added ? result : undefined;
}
function subtractQuery(id, alreadyStartedQueries, query) {
    if (!alreadyStartedQueries[id.n]) {
        alreadyStartedQueries[id.n] = {};
        alreadyStartedQueries[id.n][id.i] = makeMutableCopy(query);
        return query;
    }
    if (!alreadyStartedQueries[id.n][id.i]) {
        alreadyStartedQueries[id.n][id.i] = makeMutableCopy(query);
        return query;
    }
    return subtractQueryPart(alreadyStartedQueries[id.n][id.i], query);
}
function runQuery(type, value, query, qCtx) {
    return Promise.resolve(null)
        .then(function () { return type.id(value, qCtx.context, qCtx); })
        .then(function (id) {
        var nodeID = NodeID_1.createNodeID(type.name, id);
        var result = NodeID_1.getNode(qCtx.result, nodeID);
        var remainingQuery = subtractQuery(nodeID, qCtx.startedQueries, query);
        if (!remainingQuery) {
            return nodeID;
        }
        return Promise.all(Object.keys(remainingQuery).map(function (key) {
            var subQuery = remainingQuery[key];
            if (!(subQuery === true ||
                (subQuery != null && typeof subQuery === 'object'))) {
                result[key] = ErrorResult_1.createErrorResult('Expected subQuery to be "true" or an Object but got ' +
                    type_name_from_value_1.default(subQuery) +
                    ' while getting ' +
                    type.name +
                    '(' +
                    id +
                    ').' +
                    key, {}, 'INVALID_SUB_QUERY');
                return;
            }
            if (is_cached_1.default(qCtx.result, nodeID, key, subQuery)) {
                return;
            }
            return resolveField_1.default(type, value, key, subQuery, qCtx)
                .then(null, function (err) {
                return getErrorObject(err, 'while getting ' + type.name + '(' + nodeID.i + ').' + key, qCtx.logging);
            })
                .then(function (value) {
                result[key] = value;
            });
        })).then(function () { return nodeID; });
    }, function (err) {
        return getErrorObject(err, 'while getting ID of ' + type.name, qCtx.logging);
    });
}
exports.default = runQuery;
//# sourceMappingURL=runQuery.js.map